/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */

#include <stdio.h>
#include "cache.h"
#include "stash.h"

#include "cachectl.h"
#include "main.h"
#include "mem.h"


/* a very coarse model of TLB misses.  They just take a constant
   latency to satisfy.
 */
#define tlbpenalty 160

/* the time a cache bank is occupied due to 
   a miss when the data arrives */
#define ICACHEFILLTIME 2
#define DCACHEFILLTIME 2
#define L2CACHEFILLTIME 2
#define L3CACHEFILLTIME 8
/* the time a cache bank is occupied due to an access */
#define ICACHEACCESSTIME 1
#define DCACHEACCESSTIME 1
#define L2WBACCESSTIME 2
#define L2ACCESSTIME 1
#define L3ACCESSTIME 4
/* the time the L1/L2 bus is occupied by a data transfer */
#define BUSTRANSFERTIME 2
#define BUSWBTRANSFERTIME 2

/* miss penalties, adjusted so that the first number is what
   you think of as the latency */
#define misspenalty1 6 - ICACHEFILLTIME
#define misspenalty2 12 - L2CACHEFILLTIME
#define misspenalty3 62 - L3CACHEFILLTIME


extern unsigned long cyc;
static unsigned long miss_queue_full = 0;

long dtlb[DTLB_SIZE];
long dtlbstamp[DTLB_SIZE];
int dtlbid[DTLB_SIZE];
unsigned long dtlbavail[DTLB_SIZE];

long itlb[ITLB_SIZE];
long itlbstamp[ITLB_SIZE];
int itlbid[ITLB_SIZE];

int ihash[4096+512];
int dhash[4096+512];

unsigned long lastdaccess[16]={0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
int numiaccesses = 0;

long lastipg = 0;
int lastiid = 0;
long lastdpg = 0;
int lastdid = 0;

unsigned long lastavail = 0;

static Cent icache[IASSOC][ICACHESZ];
static Cent  dcache[DASSOC][DCACHESZ];
static Cent  scache[SASSOC][SCACHESZ],
            Tcache[TASSOC][TCACHESZ];

/* event holders for event-driven simulation of memory hierarchy */
#define HOLDERS 100
static cache_request request_holders[HOLDERS];
static cache_request *free_head;
cache_request *cache_queue_head;

/* statistics gathering for memory hierarchy */
extern unsigned long text_start_address;
static unsigned long tihit = 0, timiss = 0;
static unsigned long tdhit = 0, tdmiss = 0, tdirty = 0;
static unsigned long tread = 0, twrite = 0;
static unsigned long secmiss = 0, sechit = 0, secdirty = 0;
static unsigned long termiss = 0, terhit = 0, terdirty = 0;
unsigned long totmem=0, totmemdelay=0;

extern void initcache(void) {
    int line, index, i, j;
    tihit = timiss = tdhit = tdmiss = tdirty = tread = twrite = 0;

    for (i=0;i<HOLDERS-1;i++) {
      request_holders[i].next = &request_holders[i+1];
    }
    request_holders[99].next = NULL;
    free_head = &request_holders[0];
    cache_queue_head = NULL;

    for(line = 0; line < IASSOC; ++line)
	for(index = 0; index < ICACHESZ; ++index) {
	    icache[line][index].dirty = 0;
	    icache[line][index].touched = 0;
	}

    init_stash();

    for(line = 0; line < DASSOC; ++line)
      for(index = 0; index < DCACHESZ; ++index) {
	dcache[line][index].dirty = 0;
	dcache[line][index].touched = 0;
      }
    for (line = 0; line < 4096+512;line++) {
      ihash[line] = 0;
      dhash[line] = 0;
    }
}

static void
place_in_cache_queue(cache_request *new_entry) {
  
  cache_request *old_entry;
  cache_request *prev_entry;

  if (!cache_queue_head) {
    cache_queue_head = new_entry;
    new_entry->next = NULL;
    return;
  }
  if (new_entry->request_time < cache_queue_head->request_time) {
    new_entry->next = cache_queue_head;
    cache_queue_head = new_entry;
    return;
  }
  old_entry = cache_queue_head;
  while (old_entry->next && 
	 old_entry->next->request_time <= new_entry->request_time) {
    old_entry = old_entry->next;
  }
  new_entry->next = old_entry->next;
  old_entry->next = new_entry;
}

static void free_cache_request(cache_request *old) {

old->next = free_head;
free_head = old;
}

static long lastiaccess[ICACHEBANKS] = {0,0,0,0,0,0,0,0};
static long iacc[ICACHEBANKS], dacc[DCACHEBANKS], ifill[ICACHEBANKS],
  dfill[DCACHEBANKS], l2acc[L2CACHEBANKS], l2fill[L2CACHEBANKS],
  dwb[DCACHEBANKS], l2wb[L2CACHEBANKS], l3acc, l3fill, bus, buswb;
static long last2access[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static long last3access[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
static long lastbusaccess = 0;


static int L2hit(mem_addr tag, c_rw rw, int threadid) {

    int line, index;

    index = tag & (SCACHESZ - 1);

    for(line = 0; line < SASSOC; ++line)
      if((scache[line][index].tag == tag) &&
	 (scache[line][index].id == threadid)) {
	sechit++;
	if(rw == C_WRITE)
	  scache[line][index].dirty = 1;
	scache[line][index].touched = cyc;
	return 1;
      }
    secmiss++;
    line = 0;
    if (SASSOC != 1) {
      int leastline; long least = cyc+1;
      for(; (line < (SASSOC)); ++line) {
	if (scache[line][index].touched < least) {
	  least = scache[line][index].touched;
	  leastline = line;
	}
      }
      line = leastline;
    }
    if((scache[line][index].dirty == 1))
      secdirty++;
    scache[line][index].tag = tag;
    scache[line][index].touched = cyc;
    scache[line][index].id = threadid;
    return 0;
}

extern int L3hit(mem_addr tag, c_rw rw, int threadid) {

    int line, index;

    index = tag & (TCACHESZ - 1);

    l3acc++;

    for(line = 0; line < TASSOC; ++line)
      if((Tcache[line][index].tag == tag) &&
	 (Tcache[line][index].id == threadid)) {
	terhit++;
	if(rw == C_WRITE)
	  Tcache[line][index].dirty = 1;
	Tcache[line][index].touched = cyc;
	return 1;
      }
    termiss++;
    line = 0;
    if (TASSOC != 1) {
      int leastline; long least = cyc+1;
      for(; (line < (TASSOC)); ++line) {
	if (Tcache[line][index].touched < least) {
	  least = Tcache[line][index].touched;
	  leastline = line;
	}
      }
      line = leastline;
    }
    if((Tcache[line][index].dirty == 1))
      terdirty++;
    Tcache[line][index].tag = tag;
    Tcache[line][index].touched = cyc;
    Tcache[line][index].id = threadid;
    return 0;
}


/* this is an event-driven simulator of the cache hierarchy -- events
   are kept in chronological order, so that if there is anything to do
   this cycle, it is at the head of the queue.

   Accesses will come into the queues through doiaccess or dodaccess,
   if there is a miss,
   make several transitions, and then leave through the IFILL or DFILL
   states.

   Duplicate requests will be merged into a single request.  This 
   process is pretty complex and still prone to bugs, 
   some of which I have to admit I
   just tolerate.
*/

void
process_cache_queues() {
  cache_request *next_entry;
  int bank;
  scoreboard_t *meminst;
  context *mergecontext, *nextcontext;

  while (((next_entry = cache_queue_head) != NULL) 
	 && next_entry->request_time <= cyc) {
    cache_queue_head = next_entry->next;
    switch(next_entry->action) {
    case IFILL:
      bank = next_entry->tag & (ICACHEBANKS-1);
      ifill[bank]++;
      if (lastiaccess[bank] < cyc) {
	lastiaccess[bank] = cyc;
      }
      lastiaccess[bank] += ICACHEFILLTIME;
      if (next_entry->irequestor) {
	if (next_entry->irequestor->fetchcycle > cyc &&
	    ((next_entry->irequestor->nextpc>>LINESIZEEXP) == next_entry->tag)) {
	  next_entry->irequestor->fetchcycle = lastiaccess[bank];
	  DEBUGPRINTF("T%d can resume fetching at %ld, I miss complete\n",
		      next_entry->irequestor->id, lastiaccess[bank]);
	}
	else
	  printf("T%d I miss complete but not needed -- error\n",
		      next_entry->irequestor->id);
	mergecontext = next_entry->irequestor;
	mergecontext->cache_entry = NULL;
	while (mergecontext->mergethread) {
	  nextcontext = mergecontext->mergethread;
	  mergecontext->mergethread = NULL;
	  mergecontext = nextcontext;
	  mergecontext->cache_entry = NULL;
	  if (mergecontext->fetchcycle > cyc &&
	      ((mergecontext->nextpc)>>LINESIZEEXP == next_entry->tag)) {
	    mergecontext->fetchcycle = lastiaccess[bank];
	    DEBUGPRINTF("T%d can resume fetching at %ld, I miss complete\n",
			mergecontext->id, lastiaccess[bank]);
	    lastiaccess[bank] += ICACHEACCESSTIME;
	  }
	  else
	    DEBUGPRINTF("T%d I miss complete but not needed -- error\n",
			mergecontext->id);
	}
      }
      if (next_entry->cacheline->pending && 
	  next_entry->cacheline->tag == next_entry->tag)
	next_entry->cacheline->pending = 0;
      free_cache_request(next_entry);
      break;
    case DFILL:
      bank = next_entry->tag & (DCACHEBANKS-1);
      dfill[bank]++;
      if (lastdaccess[bank] < cyc) {
	lastdaccess[bank] = cyc;
      }
      lastdaccess[bank] += DCACHEFILLTIME;
      if (next_entry->drequestor) {
	meminst = next_entry->drequestor;
	if (meminst->fu == SYNCH)
	  synch_mem_resolve(meminst, lastdaccess[bank]);
	else
	  mem_resolve(meminst, lastdaccess[bank]);
	if (meminst->srcmem) {
	  totmem++; /* ignore merged instructions, since I don't have the
		       proper stats */
	  totmemdelay += meminst->donecycle - next_entry->start_time;
	}
	while (meminst->mergeinst) {
	  meminst = meminst->mergeinst;
	  lastdaccess[bank] += DCACHEACCESSTIME;
	  if (meminst->fu == SYNCH)
	    synch_mem_resolve(meminst, lastdaccess[bank]);
	  else
	    mem_resolve(meminst, lastdaccess[bank]);
	}
      }
      if (next_entry->cacheline->pending && 
	  next_entry->cacheline->tag == next_entry->tag)
	next_entry->cacheline->pending = 0;
      free_cache_request(next_entry);
      break;
    case DWB:
      bank = next_entry->tag & (DCACHEBANKS-1);
      dwb[bank]++;
      if (lastdaccess[bank] < cyc) {
	lastdaccess[bank] = cyc;
      }
      lastdaccess[bank] += DCACHEFILLTIME;
      next_entry->request_time = lastdaccess[bank];
      next_entry->action = BUS_WB;
      place_in_cache_queue(next_entry);
      break;
    case BUS:
      if (lastbusaccess < cyc) {
	lastbusaccess = cyc;
      }
      bus++;
      next_entry->request_time = lastbusaccess;
      lastbusaccess += BUSTRANSFERTIME;
      if (next_entry->source == ICACHE)
	next_entry->action = IFILL;
      else 
	next_entry->action = DFILL;
      place_in_cache_queue(next_entry);
      break;
    case BUS_WB:
      if (lastbusaccess < cyc) {
	lastbusaccess = cyc;
      }
      buswb++;
      lastbusaccess += BUSWBTRANSFERTIME;
      next_entry->request_time = lastbusaccess + misspenalty1;
      next_entry->action = L2_WB;
      place_in_cache_queue(next_entry);
      break;
    case L2ACCESS:
      bank = next_entry->tag & (L2CACHEBANKS-1);
      l2acc[bank]++;
      if (last2access[bank] < cyc) {
	last2access[bank] = cyc;
      }
      if (L2hit(next_entry->tag, next_entry->read_write,
		next_entry->masterid)) {
	next_entry->request_time = last2access[bank];
	next_entry->action = BUS;
      }
      else {
	next_entry->request_time = last2access[bank] + misspenalty2;
	next_entry->action = L3ACCESS;
      }
      last2access[bank] += L2ACCESSTIME;
      place_in_cache_queue(next_entry);
      break;
    case L2FILL:
      bank = next_entry->tag & (L2CACHEBANKS-1);
      l2fill[bank]++;
      if (last2access[bank] < cyc) {
	last2access[bank] = cyc;
      }
      last2access[bank] += L2CACHEFILLTIME;
      next_entry->request_time = last2access[bank];
      next_entry->action = BUS;
      place_in_cache_queue(next_entry);
      break;
    case L2_WB:
      bank = next_entry->tag & (L2CACHEBANKS-1);
      l2wb[bank]++;
      if (last2access[bank] < cyc) {
	last2access[bank] = cyc;
      }
      last2access[bank] += L2WBACCESSTIME;
      /* L1 access was a hit, so I'm just going to assume a hit */
      free_cache_request(next_entry);
      break;
    case L3ACCESS:
      bank = next_entry->tag & (L3CACHEBANKS-1);
      if (last3access[bank] < cyc) {
	last3access[bank] = cyc;
      }
      if (L3hit(next_entry->tag, next_entry->read_write,
		next_entry->masterid)) {
	next_entry->request_time = last3access[bank];
	next_entry->action = L2FILL;
      }
      else {
	next_entry->request_time = last3access[bank] + misspenalty3;
	next_entry->action = L3FILL;
      }
      last3access[bank] += L3ACCESSTIME;
      place_in_cache_queue(next_entry);
      break;
    case L3FILL:
      l3fill++;
      bank = next_entry->tag & (L3CACHEBANKS-1);
      if (last3access[bank] < cyc) {
	last3access[bank] = cyc;
      }
      last3access[bank] += L3CACHEFILLTIME;
      next_entry->request_time = last3access[bank];
      next_entry->action = L2FILL;
      place_in_cache_queue(next_entry);
      break;
    default:
      printf("ILLEGAL CACHE REQUEST\n");
      break;
    }
  }
}

static cache_request *get_c_request_holder() {

  cache_request *new;

  if (free_head) {
    new = free_head;
    free_head = new->next;
    return new;
  }
  /* didn't find one -- we don't want this ever to happen, so it is okay if
     we take drastic action.  If this happens, bump up HOLDERS */
  while (!free_head) {
    cyc++;
    /*stalling whole processor to get cache_request structures! */
    miss_queue_full++;
    process_cache_queues();
  }
  new = free_head;
  free_head = new->next;
  return new;
}

merge_irequest(mem_addr addr, context *current) {

  long tag = addr>> (LINESIZEEXP);
  cache_request *entry = cache_queue_head;
  context *entrycontext;

  while (entry && (entry->tag != tag))
    entry = entry->next;
  if (!entry)
    printf("Oops, I merge error\n");
  entrycontext = entry->irequestor;
  if (!entrycontext) {
    entry->irequestor = current;
    current->cache_entry = entry;
    return;
  }
  if (entrycontext == current)
    printf("Oops, I merge error 2\n");
  while (entrycontext->mergethread) {
    entrycontext = entrycontext->mergethread;
    if (entrycontext == current)
      printf("Oops, I merge error 3\n");
  }
  entrycontext->mergethread = current;
  current->cache_entry = entry;
  current->mergethread = NULL;
}

merge_drequest(mem_addr addr, scoreboard_t *meminst) {

  unsigned tag = addr>> (LINESIZEEXP);
  cache_request *entry = cache_queue_head;
  scoreboard_t *entryinst;

  while (entry && (entry->tag != tag || entry->source != DCACHE))
    entry = entry->next;
  if (!entry)
    printf("Oops, D merge error\n");
  entryinst = entry->drequestor;
  if (!entryinst) {
    entry->drequestor = meminst;
    meminst->mergeinst = NULL;
    return;
  }
  while (entryinst->mergeinst)
    entryinst = entryinst->mergeinst;
  entryinst->mergeinst = meminst;
  meminst->mergeinst = NULL;
}

/* brute force clean up of the cache queue after a branch mispredict */

void
clean_up_cache_queue(context *current) {

  cache_request *entry = current->cache_entry;
  context *entrycontext;
  context **pointerto;

  entrycontext = entry->irequestor;
  pointerto = &(entry->irequestor);
  while (entrycontext) {
    if (entrycontext == current) {
      *pointerto = entrycontext->mergethread;
      current->cache_entry = NULL;
      current->mergethread = NULL;
      return;
    }
    pointerto = &entrycontext->mergethread;
    entrycontext = entrycontext->mergethread;
  }
  printf("Oops, other I merge error\n");
}

extern int doiaccess(mem_addr addr, context *current)
{
  int line;
  int tag;
  int index;
  int bank;
  Cent *cline;
  int ipenalty;
  cache_request *new_request;
  
    
  tag = addr>> (LINESIZEEXP);
  index = tag & (ICACHESZ-1);
  bank = tag & (ICACHEBANKS-1);
  iacc[bank]++;
  if (lastiaccess[bank] < cyc) {
    lastiaccess[bank] = cyc;
  }
  ipenalty = lastiaccess[bank] - cyc;
  lastiaccess[bank] += ICACHEACCESSTIME;
  
  for(line = 0; line < IASSOC; ++line) {
    cline = &icache[line][index];
    if((cline->tag == tag) &&
       (cline->id == current->masterid)) {
      tihit++;
      cline->touched = cyc;
      if (cline->pending) { /* hit on a line that is not in the $ yet */
	current->fetchcycle = cyc + 100000000; /* stalls for now */
	merge_irequest(addr, current);
	return 100000000;
      }
      ipenalty += itlb_lookup(current->nextpc, current->masterid);
      current->fetchcycle = cyc + ipenalty;
      return ipenalty;
    }
  }
  
  if (IASSOC != 1) {
    int leastline; long least = cyc+1;
    for(line = 0; (line < (IASSOC)); ++line) {
      cline = &icache[line][index];
      if (cline->touched < least) {
	least = cline->touched;
	leastline = line;
      }
    }
    cline = &icache[leastline][index];
  }
  timiss++;
  cline->tag = tag;
  cline->touched = cyc;
  cline->id = current->masterid;
  cline->pending = 1;
  current->mergethread = NULL;
  /* Put a request in the cache queue */
  current->cache_entry = new_request = get_c_request_holder();
  new_request->action = L2ACCESS;
  new_request->source = ICACHE;
  new_request->read_write = C_READ;
  new_request->irequestor = current;
  new_request->address = addr;
  new_request->tag = tag;
  new_request->masterid = current->masterid;
  new_request->cacheline = cline;
  new_request->request_time = cyc + ipenalty + misspenalty1 +
    itlb_lookup(current->nextpc, current->masterid);
  current->fetchcycle += 100000000; /* stalls until access completes */
  place_in_cache_queue(new_request);
  return 100000000;
}


extern int dodaccess(mem_addr addr, c_rw rw, context *current,
		     scoreboard_t *meminst) {
    int i, line, tag, index, bank, dpenalty;
    Cent *cline;
    cache_request *new_request, *wb_request = NULL;
    int threadid = current->masterid;

    tag = addr >> (LINESIZEEXP);
    index = tag & (DCACHESZ - 1);
    bank = tag & (DCACHEBANKS-1);
    dacc[bank]++;
    if (lastdaccess[bank] < cyc) {
      lastdaccess[bank] = cyc;
    }
    dpenalty = lastdaccess[bank] - cyc;
    lastdaccess[bank] += DCACHEACCESSTIME;

    if(rw == C_READ) tread++;
    else             twrite++;
    for(line = 0; line < DASSOC; ++line) {
      cline = &dcache[line][index];
      if((cline->tag == tag) && (cline->id == threadid)) {
	tdhit++;
	if(rw == C_WRITE)
	  cline->dirty = 1;
	cline->touched = cyc;
	if (cline->pending) {
	  if (rw == C_READ) {
	    merge_drequest(addr, meminst);
	    meminst->status = MEMORY;
	    return 1;
	  }
	  else {
	    /* FIXME write to pending line */
	  }
	}
	dpenalty += dtlb_lookup(addr,threadid);
	totmem++;
	totmemdelay += dpenalty;
	return dpenalty;
      }
    }
    tdmiss++;
    line = 0;
    if (DASSOC != 1) {
      int leastline; long least = cyc+1;
      for(; (line < (DASSOC)); ++line) {
	cline = &dcache[line][index];
	if (cline->touched < least) {
	  least = cline->touched;
	  leastline = line;
	}
      }
      cline = &dcache[leastline][index];
    }
    if((cline->dirty == 1)) {
      tdirty++;
      wb_request = get_c_request_holder();
      wb_request->action = DWB;
      wb_request->source = WB;
      wb_request->tag = cline->tag;
      wb_request->request_time = cyc;
      place_in_cache_queue(wb_request);
    }
    cline->tag = tag;
    cline->touched = cyc;
    cline->id = threadid;
    cline->pending = 1;
    meminst->mergeinst = NULL;
    new_request = get_c_request_holder();
    new_request->action = L2ACCESS;
    new_request->source = DCACHE;
    new_request->read_write = rw;
    new_request->drequestor = meminst;
    new_request->address = addr;
    new_request->tag = tag;
    new_request->cacheline = cline;
    new_request->masterid = threadid;
    new_request->start_time = cyc;
    new_request->request_time = cyc + dpenalty + misspenalty1 +
      dtlb_lookup(addr, threadid);
    meminst->status = MEMORY;
    place_in_cache_queue(new_request);
    return 100000000;
}

/* ??? */
extern int cache_flush(char *addr, int nbytes, int cache) {
    int flag = -1;
    if(cache & FICACHE) {
	flag = 0;
    }
    if(cache & FDCACHE) {
	flag = 0;
    }
    return flag;
}

flush_caches(int cur) {
  int i,j;

  for (i=0;i<ICACHESZ;i++) {
    for (j=0;j<IASSOC;j++) {
      if (icache[j][i].id == cur) {
	icache[j][i].tag = 0;
      }
    }
  }
  for (i=0;i<DCACHESZ;i++) {
    for (j=0;j<DASSOC;j++) {
      if (dcache[j][i].id == cur) {
	dcache[j][i].tag = 0;
      }
    }
  }
  for (i=0;i<SCACHESZ;i++) {
    for (j=0;j<SASSOC;j++) {
      if (scache[j][i].id == cur) {
	scache[j][i].tag = 0;
      }
    }
  }
  for (i=0;i<TCACHESZ;i++) {
    for (j=0;j<TASSOC;j++) {
      if (Tcache[j][i].id == cur) {
	Tcache[j][i].tag = 0;
      }
    }
  }
  for (i=0;i<ITLB_SIZE;i++) {
    if (itlbid[i] == cur) {
      itlb[i] = 0;
      itlbstamp[i] = 0;
    }
  }
  for (i=0;i<DTLB_SIZE;i++) {
    if (dtlbid[i] == cur) {
      dtlb[i] = 0;
      dtlbstamp[i] = 0;
      dtlbavail[i] = 0;
    }
  }
}

/* ??? */
extern int cache_ctl(char *addr, int nbytes, int op) {
    int flag = -1;
    if(op & CACHEABLE) {
	flag = 0;
    }
    if(op & UNCACHEABLE) {
	flag = 0;
    }
    return flag;
}

long itlbmiss=0,dtlbmiss=0;

extern void print_cstats(void) {
  int i;

    printf("Cache Statistics\n");
    printf(" Stall for miss queue full %d cycles (%.2f%%)\n",
	   miss_queue_full, (double) miss_queue_full/cyc);
    printf(" Reads: %ld   Writes: %ld\n", tread+tihit+timiss, twrite);
    printf("ICACHE: size: %d KB assoc: %d\n", ICACHESZ*IASSOC/16, IASSOC);
    if((tihit + timiss) > 0)
	printf("ICACHE: hits:%ld misses:%ld  Hit Ratio: %.2f%%\n",
	       tihit,timiss, (double) 100 * tihit / (tihit + timiss));
    printf("DCACHE: size: %d KB assoc: %d\n", DCACHESZ*DASSOC/16, DASSOC);
    if((tdhit+tdmiss) > 0) {
	printf("DCACHE: hits:%ld misses:%ld writebacks:%d  Hit Ratio: %.2f%%\n",
	   tdhit,tdmiss,tdirty, (double) 100*tdhit/(tdhit+tdmiss));
    }
    printf("SCACHE: size: %d KB assoc: %d\n", SCACHESZ*SASSOC/16, SASSOC);
    if((sechit+secmiss) > 0) {
	printf("SCACHE: hits:%ld misses:%ld  Hit Ratio: %.2f%%\n",
	   sechit,secmiss, (double) 100*sechit/(sechit+secmiss));
    }
    printf("3CACHE: size: %d KB assoc: %d\n", TCACHESZ*TASSOC/16, TASSOC);
    if((terhit+termiss) > 0) {
	printf("3CACHE: hits:%ld misses:%ld  Hit Ratio: %.2f%%\n",
	   terhit,termiss, (double) 100*terhit/(terhit+termiss));
	printf("3CACHE: dirty items replaced:%d\n",terdirty);
    }
    printf("ITLB: size: %d, misses %d, miss rate %.2f\n",
	   ITLB_SIZE, itlbmiss, (double) 100.0*itlbmiss / (tihit+timiss));
    printf("DTLB: size: %d, misses %d, miss rate %.2f\n",
	   DTLB_SIZE, dtlbmiss, (double) 100.0*dtlbmiss / (tdhit+tdmiss));
    print_stash_stats();
  printf("avg mem delay %.3f\n", (double) totmemdelay/totmem);
  printf("icache bank util. ");
  for (i=0;i<8;i++) printf("%.3f ", (double) (ICACHEACCESSTIME*iacc[i] + 
					      ICACHEFILLTIME*ifill[i])/cyc);
  printf("\n");
  printf("dcache bank util. ");
  for (i=0;i<8;i++) printf("%.3f ",(double)(DCACHEACCESSTIME*dacc[i] + 
					    DCACHEFILLTIME*(dfill[i]+dwb[i]))/cyc);
  printf("\n");
  printf("L2 bank util. ");
  for (i=0;i<8;i++) printf("%.3f ", (double) (L2ACCESSTIME*l2acc[i] + 
					      L2CACHEFILLTIME*l2fill[i] +
					      L2WBACCESSTIME*l2wb[i])/cyc);
  printf("\n");
  printf("L1-L2 bus util. = %.3f, L3 util. = %.3f\n", 
	 (double) (bus * BUSTRANSFERTIME + buswb * BUSWBTRANSFERTIME)/cyc,
	 (double) (l3acc*L3ACCESSTIME+l3fill*L3CACHEFILLTIME)/cyc);
}


extern void zero_cstats(void) {
  int i;

  tread=tihit=timiss=twrite=0;
  tdhit=tdmiss=sechit=secmiss=0;
  terhit=termiss=itlbmiss=dtlbmiss=0;
  for (i=0;i<ICACHEBANKS;i++) {
    ifill[i] = iacc[i] = 0;
  }
  for (i=0;i<DCACHEBANKS;i++) {
    dacc[i] = dfill[i] = l2fill[i] = l2acc[i] = 0;
  }
  bus = 0;
  l3fill = l3acc = 0;
}

/* TLB routines */

int itlb_lookup(long address, int id) {

  long pg = address >> PGSIZE;
  int i;
  int penalty, firsttry;
  int least;
  long earliest;

  /*try to avoid the search */
  if (pg == lastipg && id == lastiid ) return 0; 
  lastipg = pg;
  lastiid = id;
  firsttry = ihash[(pg & 511) + (id << 8)];
  if ( pg == itlb[firsttry] && id == itlbid[firsttry]) {
    itlbstamp[firsttry] = cyc;
    return 0;
  }
  for (i=0;i<ITLB_SIZE;i++) {
    if (pg == itlb[i] && id == itlbid[i]) {
      itlbstamp[i] = cyc;
      return 0;
    }
  }
  itlbmiss++;
  earliest = cyc + 1;
  for (i=0;i<ITLB_SIZE;i++) { /*find lru page*/
    if (itlbstamp[i] < earliest) {
      least = i;
      earliest = itlbstamp[i];
    }
  }
  itlb[least] = pg;
  itlbstamp[least] = cyc;
  itlbid[least] = id;
  ihash[(pg & 511) + (id << 8)] = least;
  return tlbpenalty;
}

int dtlb_lookup(long address, int id) {

  long pg = address >> PGSIZE;
  int i;
  int penalty, firsttry;
  int least;
  long earliest;

  if (cyc >= lastavail) {
    if (pg == lastdpg && id == lastdid) return 0;
  }
  else {
    if (pg == lastdpg && id == lastdid) return (lastavail - cyc);
  }
  lastdpg = pg;
  lastdid = id;
  firsttry = dhash[(pg & 511) + (id << 8)];
  if (pg == dtlb[firsttry] && id == dtlbid[firsttry]) {
    dtlbstamp[firsttry] = cyc;
    lastavail = dtlbavail[firsttry];
    if (cyc >= lastavail)
      return 0;
    else
      return lastavail - cyc;
  }
  for (i=0;i<DTLB_SIZE;i++) {
    if (pg == dtlb[i] && id == dtlbid[i] ) {
      dtlbstamp[i] = cyc;
      lastavail = dtlbavail[i];
      if (cyc >= lastavail)
	return 0;
      else
	return lastavail - cyc;
    }
  }
  dtlbmiss++;
  earliest = cyc + 1;
  for (i=0;i<DTLB_SIZE;i++) { /*find lru page*/
    if (dtlbstamp[i] < earliest) {
      least = i;
      earliest = dtlbstamp[i];
    }
  }
  /*flush pages from cache??? */
  dtlb[least] = pg;
  dtlbstamp[least] = cyc;
  dtlbid[least] = id;
  dtlbavail[least] = cyc + tlbpenalty;
  lastavail = cyc + tlbpenalty;
  dhash[(pg & 511) + (id << 8) ] = least;
  return tlbpenalty;
}

init_tlbs() {
  int i;
  for (i=0;i<ITLB_SIZE;i++) 
    itlbstamp[i] = -1;
  for (i=0;i<DTLB_SIZE;i++)
    dtlbstamp[i] = -1;
}
