/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Jack Lo.
Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
 */


#include "main.h"
#include "commit.h"
#include "inst.h"
#include "smt.h"

extern int numcontexts;
extern context *thecontexts[];
extern int i_registers_used;
extern int f_registers_used;
extern long allinstructions;
extern unsigned long cyc;
extern scoreboard_t *lock_failed[];
extern acq_entry *acquire_queue[];
extern int debugsync;

int i_registers_freed = 0;
int f_registers_freed = 0;

/* We record the total number of renaming registers freed
   while we're committing instructions, but we don't want 
   to actually make them available until the end of the cycle,
   which is when this is called.
*/

fix_regs() {
  i_registers_used -= i_registers_freed;
  i_registers_freed = 0;
  f_registers_used -= f_registers_freed;
  f_registers_freed = 0;
}

/* This is the final stage.  A number of things get done here.
   Instructions that have reached this stage commit in order 
   (per thread)..
   We are no longer keeping track of instructions through the
   pipeline (since they are no longer in order), we're just 
   going through the scoreboard (should really be called the
   active list or reorder buffer), which is in-order, and 
   seeing who is ready to commit.

   Here, renaming registers are freed.  Synch operations
   actually get performed and become visible to other
   threads.  Prediction tables get updates.
 */

void
commit()
{
  int i,j,next;
  context *current;
  scoreboard_t *top;
  unsigned long va;

  for (i=0;i<numcontexts;i++) {
    current = thecontexts[i];
    /*retire instructions */
    next = current->next_to_commit;
    top = &current->score[next];
    while ((top->status & SQUASHED) ||
           (top->status & RETIREABLE && top->donecycle <= cyc - 3)) {
      if (top->fu == SYNCH && !(top->status & SQUASHED) && 
	  !(top->syncop == LDL_L || top->syncop == LDQ_L)) {
	synchexecute(top);
      }
      if (top->status & RETIREABLE) {
	allinstructions--;
	if (top->condbranch) {
	  update_pht(top->pc,top->taken,top->ghr);
	}
      }
      else if ((top->syncop == SMT_RELEASE || top->syncop == WRITE_FILL) &&
	       thecontexts[top->thread]->early_release_inst == top->id) {
	/* abort the speculative release -- this flushes the improper thread
	   from the machine */
	context *dead_thread = thecontexts[top->thread];
#ifdef DEBUG
	  if (debugsync)
	    printf("Speculative release of context %d by context %d aborted\n",
		   dead_thread->early_release_thread,top->thread);
#endif
	if (top->syncop == SMT_RELEASE) {
	  lock_failed[dead_thread->early_release_thread] = 
	    acquire_queue[dead_thread->early_release_thread]->inst;
	}
	else {
	  lock_failed[dead_thread->early_release_thread] = 
	    thecontexts[dead_thread->early_release_thread]->wait_on_full[dead_thread->early_release_reg];
	}	  
	dead_thread->early_release_inst = NONE;
	dead_thread->early_release_thread = NONE;
	dead_thread->early_release_reg = NONE;
      }
      top->status = INVALID;
      i_registers_freed += top->iregs_used;
      f_registers_freed += top->fregs_used;
      DEBUGPRINTF("T%d, slot %d retired, %d reg freed\n",current->id,
		  next, top->iregs_used + top->fregs_used);
      next = (next + 1) & (SCORESIZE - 1);
      if (lock_failed[i])
	break;
      top = &current->score[next];
    }
    current->next_to_commit = next;
  }
}

