/* SMTSIM simulator.
   
   Copyright (C) 1994-1999 by Dean Tullsen (tullsen@cs.ucsd.edu)
   ALL RIGHTS RESERVED.

   SMTSIM is distributed under the following conditions:

     You may make copies of SMTSIM for your own use and modify those copies.

     All copies of SMTSIM must retain all copyright notices contained within.

     You may not sell SMTSIM or distribute SMTSIM in conjunction with a
     commerical product or service without the express written consent of
     Dean Tullsen.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.

Significant parts of the SMTSIM simulator were written by Gun Sirer
(before it became the SMTSIM simulator) and by Jack Lo (after it became
the SMTSIM simulator).  Therefore the following copyrights may also apply:

Copyright (C) Jack Lo
Copyright (C) E. Gun Sirer

Pieces of this code may have been derived from Jim Larus' SPIM simulator,
which contains the following copyright:

==============================================================
   Copyright (C) 1990-1998 by James Larus (larus@cs.wisc.edu).
   ALL RIGHTS RESERVED.

   SPIM is distributed under the following conditions:

     You may make copies of SPIM for your own use and modify those copies.

     All copies of SPIM must retain my name and copyright notice.

     You may not sell SPIM or distributed SPIM in conjunction with a
     commerical product or service without the expressed written consent of
     James Larus.

   THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
   IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
   WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   PURPOSE.
===============================================================
 */


#include <stdio.h>
#include <assert.h>
#include "stash.h"
#include "main.h"
#include "cache.h"
#include "predict.h"
#include "fetch.h"
#include "smt.h"
#include "inst.h"

extern int numcontexts;
extern context *thecontexts[];
extern unsigned long cyc;
extern int simulating;
extern int spec_release;
extern int spec_fill;
extern scoreboard_t *lock_failed[];
extern scoreboard_t *mispredict_discovered[];

int priorityslot;
int priority[MAXTHREADS];
int key[MAXTHREADS];

unsigned long scoreconf = 0;

static int ibank;

extern scoreboard_t *decode_head;
extern scoreboard_t *decode_tail;

extern int num_in_decode;

unsigned long misfetchtotal=0;

/* for each thread, emulate some number of instructions
 *    before beginning actual simulation.
 */

fast_forward(long ffdist[]) {

  int i;
  long j;
  StashData *stash;
  scoreboard_t *top;
  context *current;

  for (i=0;i<numcontexts;i++) {
    current = thecontexts[i];
    for (j=0;j<ffdist[i];j++) {
      current->pc = current->nextpc;
      current->nextpc += 4;

      if (!get_stash_data((mem_addr)current->pc, current->id, &stash)) {
	decode_inst(current, stash);
      }
      (*stash->emulate)(current,stash);
      if (stash->condbranch && stash->taken) {
	current->nextpc = stash->br_target;
      }
    }
  }
}
  
static void insert_decode_queue(scoreboard_t *top) 
{
  top->next = NULL;
  if (decode_head == NULL) {
    decode_head = decode_tail = top;
  }
  else {
    decode_tail->next = top;
    decode_tail = top;
  }
}


/* decide on the next thread that has the opportunity to fetch this cycle.
 */

static context *
get_next_thread()
{
	int found = 0;
	context *current;

	while((priorityslot < numcontexts))
	  {
	    current = thecontexts[priority[priorityslot++]];
	    
	    /* this thread ok if it is running, not waiting for a cache miss,
	       not blocked for some reason, and will not cause a bank conflict
	       with a thread already accessed this cycle */
	    
	    if(	(current->running) && (!current->blocked) &&
		(current->fetchcycle <= cyc) &&
		!(ibank & (1<<((current->nextpc >> LINESIZEEXP) & (ICACHEBANKS - 1)))) )
	      {
		current->fthiscycle = 0;
		ibank |= 1 << ((current->nextpc >> LINESIZEEXP) & (ICACHEBANKS - 1));
		return current;
	      }
	  }
	
	/* didn't find anything */
	return (context *) 0;
}


/* fetch the next instruction for the current thread.*/

static int
get_next_instruction(context *current)
{
  int index;
  
  
  if(	!(current->running) )
    {
      /*
	this should only happen if recent instructions caused this
	*/
      return(0);
    }
  
  /* have we crossed a cache line boundary? */
  if (current->fthiscycle && 
      ((current->nextpc != current->pc + 4) ||
      ((current->nextpc & FBLOCKMASK) != (current->pc & FBLOCKMASK))))
    return 0;
  
  /* do we have the data structures available to handle this? */
  index = (current->scoretop + 1) & (SCORESIZE - 1);
  if (!(current->score[index].status & INVALID) || 
      (&current->score[index] == lock_failed[current->id]) ||
      (&current->score[index] == mispredict_discovered[current->id]))
    {
      DEBUGPRINTF("uh-oh.  Out of scoreboard entries!\n");
      scoreconf++;
      return(0);
    }
  current->scoretop = index;
  return(1);
}


/*  This routines collects all the static information about a particular
       instruction--it's dependencies, it's earliest execution time,
       it's control dependencies, it's branch behavior, and much more.
 */

static int
setup_instruction(context *current, StashData *stash)
{
   int src1,src2;
   int writer;
   scoreboard_t *top;
   long index, j;

   top = &current->score[current->scoretop];

/* don't send nops through the pipeline */
   if (	!stash->branch &&
	!stash->smt_primitive &&
	((stash->desthaz == 31 && (stash->whichfu == INTEGER ||
		       (stash->whichfu == INTLDST && !current->destmem)) || 
	  (stash->desthaz == 63 && stash->whichfu == FP))))
     {
       /*
       ** If destination register for non-branch / non-SMT-synch
       ** instuctions is 31 (integer) or 63 (floating point),
       ** then treat as a no-op since these registers are
       ** always zero.
       */
       top->noop = 1;
       current->srcmem = 0;
       current->scoretop = (current->scoretop - 1) & (SCORESIZE - 1);
       return 0;
     }
   
#ifdef PRIO_INST_COUNT
	  key[current->id]++;
#else
#ifdef PRIO_BR_COUNT
	  if (stash->branch || stash->cond_branch) {
	    key[current->id]++;
	  }
#endif
#endif

/* A misfetch is a BTB miss that will be fixed (and predicted
    correctly) when the branch hits the decode stage.  We treat
    instructions after a misfetch differently than instructions
    after a mispredict, because misfetched instructions never
    reach the execute, or even issue, stage.  Therefore, we
    don't really care what they are.
 */

   if (current->misfetching) {
     misfetchtotal++;
     current->num_misfetches++;
     current->scoretop = (current->scoretop - 1) & (SCORESIZE - 1);
     return 0;
   }
   
	top->noop = 0;

	src1 = stash->srchaz1;
	src2 = stash->srchaz2;

	/* Copy stash into scoreboard */
	top->fu = stash->whichfu;
	top->src1 = src1;
	top->src2 = src2;
	top->dest = stash->desthaz;
	top->srcmem = current->srcmem;
	top->destmem = current->destmem;
	current->srcmem = 0;
	current->destmem = 0;
	top->donecycle = -1;
	top->iregs_used = 0;
	top->fregs_used = 0;
	top->status = FETCHED;
	top->branch = stash->branch;
	top->taken = stash->taken;
	top->condbranch = stash->condbranch;
	top->mispredict = 0;
        top->misfetch = 0;
	top->syncop = stash->syncop;
	top->deps = 0;
	top->wait_sync = 0;
	top->delay = stash->delay;
	top->thread = current->id;
	top->pc = current->pc;

	if (current->wrong_path) {
		top->wp = 1;
		current->wpinstrs++;
	}
	else {
		top->wp = 0;
		current->instrs++;
	}
        if (top->branch) {
	  if (top->condbranch) {
	    if (current->wrong_path || current->follow_sync) {
	      if ( get_bpredict(current->pc, current->ghr) ) {
		current->nextpc = stash->br_target;
		top->taken = 1;
	      }
	      else top->taken = 0;
	    }
	    else {
	      if (top->taken) {
		current->nextpc = stash->br_target;
	      }
	      current->misfetch = 1 - btblookup(current->pc, 0,0,
						current->masterid,
						stash->taken, BRANCH);
	      current->prediction = bpredict(current->pc, stash->taken,
					     current->ghr);
	      if (current->prediction == PREDICT_WRONG && simulating) {
		/*mispredicted branch */
		current->rightpc = current->nextpc;
		if (top->taken)
		  current->nextpc = current->pc + 4;
		else
		  current->nextpc =  stash->br_target;
	      }
	      else if (current->misfetch && simulating) {
		current->rightpc = current->nextpc;
		current->nextpc = current->pc + 4;
	      }
	    }
	    top->ghr = current->ghr;
	    current->ghr = ((current->ghr << 1) | top->taken) & PHTMASK; 
	  }
	  if (current->prediction == PREDICT_WRONG) {
	    DEBUGPRINTF("T%d mispredicting\n", current->id);
	    top->mispredict = 1;
	    current->misfetch = 0;
	    current->wrong_path = 1;
	    current->prediction = PREDICT_CORRECT;
	    /* Checkpoint for later recovery from mispredict. */
	    bcopy(current->lasthazard, current->lasthazard_cp,MAXREG * sizeof(long));
	    bcopy(current->writer, current->writer_cp, MAXREG * sizeof(int));
	    bcopy(current->R, current->int_reg_cp,32 * sizeof(long));
	    bcopy(current->FPR, current->fp_reg_cp,32 * sizeof(double));
	    bcopy(current->return_stack, current->return_stack_cp,
		  12 * sizeof(long));
	    current->rs_size_cp = current->rs_size;
	    current->rs_start_cp = current->rs_start;
	    current->ghr_cp = current->ghr;
	  }
	  else	if (current->misfetch) {
	    DEBUGPRINTF("T%d misfetching\n", current->id);
	    top->misfetch = 1;
	    current->misfetch = 0;
	    current->misfetching = 1;
	    current->wrong_path = 1;
	  }
	}
	if (current->prediction == PREDICT_WRONG)
	  printf("OOPS\n");
	
	/* Handle SRC 1 dependencies (INT reg 31/FLOAT reg 63 tied to zero) */
        if (src1 != 31 && src1 != 63) {
	  writer = current->writer[src1];
	  if (writer != NONE) {
	    top->deps++;
	    top->status |= BLOCKED;
	    current->score[writer].waiter[current->score[writer].numwaiting] = top;
	    current->score[writer].numwaiting++;
	  }
	}
 

	/* Handle SRC 2 dependencies (INT reg 31/FLOAT reg 63 tied to zero) */
      if (src2 != 31 && src2 != 63 && (src1 != src2)) {
	writer = current->writer[src2];
	if (writer != NONE) {
	  top->deps++;
	  top->status |= BLOCKED;
	  current->score[writer].waiter[current->score[writer].numwaiting] = top;
	  current->score[writer].numwaiting++;
	}
      }


	/* If not blocked, when ready?
	 -readycycle is nonzero only when there are no more dependencies
	 with in-flight instructions
	 */
      if (!(top->status & BLOCKED) && current->fetchcycle <= cyc) {
	top->readycycle = cyc; 
	if (src1 != 31) {
	  if ((j=current->lasthazard[src1]) > cyc) {
	    top->readycycle = j;
	  }
	}
	if (src2 != 31) {
	  if ((j=current->lasthazard[src2]) > top->readycycle) {
	    top->readycycle = j;
	  }
	}
      }
      else {
	/* top->readycycle being zero implies the instruction is not ready
	   to execute.  This was an unfortunate choice, but now it is all
	   through the code.  
	   */
	top->readycycle = 0;
	top->hazcycle = cyc;
      }
      if (top->dest < 31) {
	current->writer[top->dest] = current->scoretop;
      }
      if (top->dest > 31 && top->dest != 63) {
	current->writer[top->dest] = current->scoretop;
      }
      if (current->follow_sync) {
	if ((top->fu == INTLDST || top->fu == SYNCH))
	  top->wait_sync = 1;
      }
   /* instructions following certain synch instructions must be handled
      differently, because they are truly speculative (that is,
      even the simulator doesn't know if they will commit or not.
      */
      if (current->nextsync == NONE && !current->wrong_path) {
	if (top->syncop == SMT_HW_LOCK || top->syncop == LDL_L
	    || top->syncop == LDQ_L || top->syncop == STL_C
	    || top->syncop == STQ_C || top->syncop == READ_FULL
	    || top->syncop == WRITE_FILL) {
	  current->nextsync = top->id;
	  current->follow_sync = 1;
	  
	  DEBUGPRINTF("T%d in follow_sync mode\n", current->id);
	  bcopy(current->lasthazard, current->lasthazard_sync_cp,MAXREG * sizeof(long));
	  bcopy(current->writer, current->writer_sync_cp, MAXREG * sizeof(int));
	  bcopy(current->R, current->int_reg_sync_cp,32 * sizeof(long));
	  bcopy(current->FPR, current->fp_reg_sync_cp,32 * sizeof(double));
	  bcopy(current->return_stack, current->return_stack_sync_cp,
		12 * sizeof(long));
	  current->rs_size_sync_cp = current->rs_size;
	  current->rs_start_sync_cp = current->rs_start;
	  current->ghr_sync_cp = current->ghr;
	  current->restore_point = top->id;
	}
      }
      if (top->syncop == WRITE_FILL) {
	top->fer_thread = current->R[src2];
	top->fer_value = current->FPR[src1-32];
	top->fer_reg = src1;
      }
      if (((top->syncop == WRITE_FILL && spec_fill) || 
	  (top->syncop == SMT_RELEASE && spec_release)) &&
	  !current->wrong_path && current->early_release_thread == NONE)
	check_early_release(top);
    return 1;
 }


/* The key[] structure is changed at various places in the code.
   This mechanism implements the ICOUNT fetch mechanism from 
   the ISCA96 smt paper.
   */

calculate_priority()
{
  int changing = 1, temp, i, j;
  for (j=1;j<numcontexts && changing;j++) {
    changing = 0;
    for (i=0;i<numcontexts-j;i++) {
      if (key[priority[i]]>key[priority[i+1]]) {
	changing = 1;
	temp = priority[i];priority[i]=priority[i+1];priority[i+1]=temp;
      }
    }
  }
}

/*  Simulates the fetching of blocks of instructions. */

void
fetch()
{
  int threadsfetched = 0;
  int totalfetch = 0;
  int imemdelay;
  int itlbdelay;
  StashData *stash;
  scoreboard_t *top;
  context *current;

  /* if decode stage clear, move fetched instructions to the
     decode stage
     */
  if (num_in_decode != 0) {
    DEBUGPRINTF("Fetch backed up\n");
    return;
  }
  ibank = 0;

  while(threadsfetched++ < THREADFETCHLIMIT && (current = get_next_thread()))
    {
      /* here I'm assuming that an I fetch miss is buffered for future access,
	 making the next fetch succeed without an I cache access.  This
	 is optimistic, but avoids some nasty race conditions.
	 */
      if (current->wait_for_fetch)
	current->wait_for_fetch = 0;
      else {
	imemdelay = doiaccess((mem_addr) current->nextpc, current);
	if (imemdelay) {
	  if (imemdelay == 10000) {
	    DEBUGPRINTF("T%d icache or itlb delay\n", current->id);
	  }
	  else {
	    DEBUGPRINTF("T%d icache conflict %d cycles\n", current->id,
			imemdelay);
	  }
	  current->wait_for_fetch = 1;
	  continue;
	}
      }

      current->fthiscycle = 0;
      
      while((totalfetch < TOTALFETCHLIMIT) && 
	    (current->fthiscycle < FETCHLIMIT) &&
	    get_next_instruction(current))
	{
	  current->pc = current->nextpc;
	  current->nextpc += 4;
	  SmtDISASSEMBLE(current->id, current->pc, 1);
	  /* only decode an instruction for which we do not already
	     have the decode info */
	  if (!get_stash_data((mem_addr)current->pc, current->id, &stash))
	    decode_inst(current, stash);
	  /* instructions are emulated in the fetch stage instead of,
	     for example, the execute stage, just because it is easier
	     to do it in an in-order part of the pipeline. */
	  if (!current->misfetching)
	    (*stash->emulate)(current,stash);
	  if (setup_instruction(current, stash)) /*setup_instruction returns
						   zero for a no-op */
	    {
	      insert_decode_queue(&current->score[current->scoretop]);
	      num_in_decode++;
	    }
	  totalfetch++;
	  current->fthiscycle++;
	}
    }
  priorityslot = 0;
}
